# frozen_string_literal: true

module WAG
  class Module
    include WAG::Encodable

    def initialize
      @data = []
      @elements = []
      @exports = []
      @functions = []
      @globals = []
      @imports = []
      @memories = []
      @tables = []
      @types = []
    end

    def build(&block)
      instance_exec(&block)
      self
    end

    def import(module_name, name, &block)
      import = Import.new(module_name, name)
      @imports << import
      import.instance_exec(&block) if block
      import
    end

    def export(name, &block)
      export = Export.new(name)
      @exports << export
      export.instance_exec(&block) if block
      export
    end

    def memory(number, min = nil, max = nil, &block)
      mem = Memory.new(number, min, max)
      @memories << mem
      mem.instance_exec(&block) if block
      mem
    end

    def global(label, type, &block)
      global = Global.new(label, type)
      @globals << global
      global.instance_exec(&block) if block
      global
    end

    def type(label, &block)
      type = FunctionType.new(label)
      @types << type
      type.instance_exec(&block) if block
      type
    end

    def elem(table_id, *labels, &block)
      elem = Element.new(table_id, *labels)
      @elements << elem
      elem.instance_exec(&block) if block
      elem
    end

    def data(offset, value)
      data = Data.new(offset, value)
      @data << data
      data
    end

    def func(label = nil, &block)
      func = Function.new(label)
      @functions << func
      func.instance_exec(&block) if block
      func
    end

    def table(elements, type = :anyfunc, &block)
      table = Table.new(elements, type)
      @tables << table
      table.instance_exec(&block) if block
      table
    end

    def to_sexpr # rubocop:disable Metrics/AbcSize, Metrics/MethodLength
      mod = [:module]
      mod.concat(@imports.map(&:to_sexpr))
      mod.concat(@exports.map(&:to_sexpr))
      mod.concat(@memories.map(&:to_sexpr))
      mod.concat(@globals.map(&:to_sexpr))
      mod.concat(@types.map(&:to_sexpr))
      mod.concat(@tables.map(&:to_sexpr))
      mod.concat(@elements.map(&:to_sexpr))
      mod.concat(@data.map(&:to_sexpr))
      mod.concat(@functions.map(&:to_sexpr))
      mod
    end
  end
end
